# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright 2019 Sanhe Hu <https://github.com/MacHu-GWU/configirl-project>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is a python config management tool to manage config parameter in
centralized place. The purpose of this tool is to avoid maintain complex
config/paramater handling logic in shell script, cloudformation, terraform
and any other devops tools. Instead, we manage that in Python.

Since Python is a full featured general programming language and it is
available on any Mac / Linux machine.

It allows different DevOps tools to easily talk to each other via JSON.

This library implemented in pure Python with no dependencies.
"""

from __future__ import print_function

__version__ = "0.0.10"
__short_description__ = "Centralized Config Management Tool."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"

import os
import re
import sys
import json
import copy
import string
import inspect
from collections import OrderedDict
import argparse
import importlib

if sys.version_info.major >= 3 and sys.version_info.minor >= 5:  # pragma: no cover
    import typing


def strip_comment_line_with_symbol(line, start):
    """
    Strip comments from line string.
    """
    parts = line.split(start)
    counts = [len(re.findall(r'(?:^|[^"\\]|(?:\\\\|\\")+)(")', part))
              for part in parts]
    total = 0
    for nr, count in enumerate(counts):
        total += count
        if total % 2 == 0:
            return start.join(parts[:nr + 1]).rstrip()
    else:  # pragma: no cover
        return line.rstrip()


def strip_comments(string, comment_symbols=frozenset(('#', '//'))):
    """
    Strip comments from json string.

    :param string: A string containing json with comments started by comment_symbols.
    :param comment_symbols: Iterable of symbols that start a line comment (default # or //).
    :return: The string with the comments removed.
    """
    lines = string.splitlines()
    for k in range(len(lines)):
        for symbol in comment_symbols:
            lines[k] = strip_comment_line_with_symbol(lines[k], start=symbol)
    return '\n'.join(lines)


def read_text(abspath, encoding="utf-8"):
    """
    Read string from a file.

    :type abspath: str
    :type encoding: str
    :rtype: str
    """
    with open(abspath, "rb") as f:
        return f.read().decode(encoding)


def write_text(text, abspath, encoding="utf-8"):
    """
    Write string to a file.

    :type text: str
    :type abspath: str
    :type encoding: str
    """
    with open(abspath, "wb") as f:
        return f.write(text.encode(encoding))


def json_loads(text):
    """
    Load data from json, ignoring comments.

    :rtype: dict
    """
    return json.loads(strip_comments(text))


def json_dumps(data):
    """
    Dump data to string.

    :rtype: str
    """
    return json.dumps(data, indent=4, sort_keys=False, ensure_ascii=False)


def json_load(path):  # pragma: no cover
    """
    Load data from a json file, ignoring comments.

    :type path: str
    """
    with open(path, "rb") as f:
        return json_loads(f.read().decode("utf-8"))


def json_dump(data, path, overwrite=False):  # pragma: no cover
    """
    Dump data to a json file.

    :type data: dict
    :type path: str
    :type overwrite: bool
    """
    if not overwrite:
        if os.path.exists(path):
            raise EnvironmentError("%s already exists!" % path)
    with open(path, "wb") as f:
        f.write(json_dumps(data).encode("utf-8"))


def add_metaclass(metaclass):  # pragma: no cover
    """
    Class decorator for creating a class with a metaclass.

    This method is copied from six.py
    """

    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        if hasattr(cls, '__qualname__'):
            orig_vars['__qualname__'] = cls.__qualname__
        return metaclass(cls.__name__, cls.__bases__, orig_vars)

    return wrapper


try:
    # python3 renamed copy_reg to copyreg
    import copyreg
except ImportError:  # pragma: no cover
    import copy_reg as copyreg


class Sentinel(object):  # pragma: no cover
    _existing_instances = {}

    def __init__(self, name):
        super(Sentinel, self).__init__()
        self._name = name
        self._existing_instances[self._name] = self

    def __repr__(self):
        return "<{0}>".format(self._name)

    def __getnewargs__(self):
        return (self._name,)

    def __new__(cls, name, obj_id=None):  # obj_id is for compatibility with previous versions
        existing_instance = cls._existing_instances.get(name)
        if existing_instance is not None:
            return existing_instance
        return super(Sentinel, cls).__new__(cls)


# obj_id is for compat. with prev. versions
def _sentinel_unpickler(name, obj_id=None):
    if name in Sentinel._existing_instances:
        return Sentinel._existing_instances[name]
    return Sentinel(name)


def _sentinel_pickler(sentinel):
    return _sentinel_unpickler, sentinel.__getnewargs__()


copyreg.pickle(Sentinel, _sentinel_pickler, _sentinel_unpickler)

NOTHING = Sentinel("NOTHING")
REQUIRED = Sentinel("REQUIRED")
OPTIONAL = Sentinel("OPTIONAL")


# --- Exception Class
class ValueNotSetError(Exception):
    """
    Raises when trying to get value of a field that have not set value before.
    """
    pass


class DerivableSetValueError(Exception):
    """
    Raises when trying to set value for Derivable Field.
    """
    pass


# --- Field Class
class Field(object):
    """
    Base class for config value field.

    :type default: typing.Any
    :param default: default value for this field.

    :type dont_dump: bool
    :param dont_dump: if true, then you can't get the value if ``check_dont_dump = True``
        in :meth:`BaseConfigClass.to_dict` and :meth:`BaseConfigClass.to_json`.
        **this prevent from writing sensitive information to file**.

    :type printable: bool
    :param printable: if False, then it will not be displayed with print function.
        **this prevent from displaying sensitive information to the console**.

    :type cache: bool
    :param cache: A flag indicates that whether the cache is enabled.
        only available for :class:`Derivable` if True,
        **then it will cache computation expensive derived value**.
    """
    _creation_index = 0

    def __init__(self,
                 default=NOTHING,
                 dont_dump=False,
                 printable=True,
                 cache=False):
        # name will be update in ConfigMeta meta class.
        self.name = NOTHING
        if callable(default):
            self._value = default()
        else:
            self._value = default
        self.dont_dump = dont_dump  # type: bool
        self.printable = printable  # type: bool
        self.cache = cache  # type: bool

        # _config_object stores the Config object associated with this field
        self._config_object = NOTHING  # type: BaseConfigClass
        # _creation_index used to sort declared fields
        self._creation_index = Field._creation_index  # type: int
        Field._creation_index += 1
        # decorator
        self._getter_method = NOTHING  # type: callable

    def __repr__(self):
        return "{}(name={!r}, value={!r})".format(self.__class__.__name__, self.name, self._value)

    def set_value(self, value):
        """
        An abstract method that set value to this field.
        """
        raise NotImplementedError

    def _get_value(self, **kwargs):
        """
        An abstract method that defines get-value logic flow.
        """
        raise NotImplementedError

    def get_value(self,
                  check_dont_dump=False,
                  check_printable=False,
                  **kwargs):
        """
        Returns the value for this field.

        :type check_dont_dump: bool
        :param check_dont_dump:

        :type check_printable: bool
        :param check_printable:
        :return:

        **CN Doc**

        对于 Constant Field:

        - 如果: self.value = NOTHING, 同时 .set_value(...) 方法从来没有被调用过.
        - 如果: self.value 不等于 NOTHING, 说明 .set_value(...) 方法被吊用过, 则
            返回 self.value

        对于 Derivable Field:

        - 如果: self._getter_method() 没有成功
        """
        if self._config_object is NOTHING:
            raise AttributeError("Field.get_value() can't be called without "
                                 "initialized.")
        if check_dont_dump:
            if self.dont_dump:
                raise DontDumpError(
                    "doesn't allow to dump `{}` field".format(self.name))
        if check_printable:
            if not self.printable:
                return "***HIDDEN***"

        return self._get_value(**kwargs)

    def get_value_from_env(self, prefix=""):
        """
        Use config value stored in environment variables. This usually used
        for computation server that doesn't come with the config file. Since
        config file with sensitive information may not easy to manage. A common
        use case is AWS Lambda Function.

        :param prefix: a prefix append left to the config field name. For exmaple,
            if the config field is ``PROJECT_NAME``, and the prefix is ``MY_PROJECT_``,
            then it will read value from ``MY_PROJECT_PROJECT_NAME``.
        """
        return os.environ[prefix + self.name]

    def get_value_for_lbd(self, prefix=""):
        """
        Smartly decide where should read config value from.
        """
        if self._config_object.is_aws_lambda_runtime():
            return self.get_value_from_env(prefix=prefix)
        else:
            return self.get_value()

    def _validate_method(self, config_object, value):
        return True

    def validator(self, method):
        """
        A decorator to bind validate method.

        :type method: callable
        :param method: a callable function like ``method(self, value)``
            that take ``self`` as first parameters representing the config object.
            ``value`` as second parameters to represent the value you want to validate.
        """
        self._validate_method = method

    def validate(self, *args, **kwargs):
        """
        An abstract method executes the validator method.
        """
        self._validate_method(self._config_object, self.get_value())


class DontDumpError(Exception):
    """
    Raises when trying to dump a ``dont_dump=True`` config value.
    """
    pass


class Constant(Field):
    """
    Constant Value Field.
    """

    def set_value(self, value):
        """
        Set value to this Constant field.
        """
        self._value = value

    def _get_value(self, **kwargs):
        """
        Constant field get-value logic flow.

        If ``self._value`` is ``NOTHING``, it meas the .set_value(...) method
        never be called. Otherwise, return the ``self._value``
        """
        if self._value is NOTHING:
            raise ValueNotSetError(
                "{}.{} has not set a value yet!".format(
                    self._config_object.__class__.__name__, self.name
                )
            )
        return self._value


class Derivable(Field):
    """
    Derivable Value Field.
    """

    def set_value(self, value):
        """
        Derivable field doesn't allow to set value manually!
        """
        raise DerivableSetValueError(
            "{} is a Derivable field, you cannot set value to it".format(
                self.name
            ))

    def getter(self, method):
        """
        A decorator to bind getter method.
        """
        self._getter_method = method

    def _get_value(self, **kwargs):
        """
        Derivable field get-value logic flow.

        If ``self._getter_method`` is ``NOTHING``, it means the getter method
        is not implemented yet.

        If cache is enabled, try to retrieve value from existing value stored in
        ``self._value``. If not available, call the getter method to retrieve
        the value.

        If cache is disabled, call the getter method.
        """
        if self._getter_method is NOTHING:
            raise NotImplementedError(
                "{}.{} getter method is not implemented, "
                "use @{}.getter to decorate a getter function.".format(
                    self._config_object.__class__.__name__, self.name, self.name
                ))

        try:
            if self.cache:
                if self._value is NOTHING:
                    self._value = self._getter_method(
                        self._config_object, **kwargs)
                return self._value
            else:
                return self._getter_method(self._config_object, **kwargs)
        except ValueNotSetError as e:  # dependent constant value not set yet
            raise ValueNotSetError(
                "can't get {}.{}, because: {}".format(
                    self._config_object.__class__.__name__, self.name, e
                )
            )
        except Exception as e:
            raise e


def is_instance_or_subclass(val, class_):
    """Return True if ``val`` is either a subclass or instance of ``class_``."""
    try:
        return issubclass(val, class_)
    except TypeError:
        return isinstance(val, class_)


def _get_fields(attrs, field_class, pop=False, ordered=False):
    """Get fields from a class. If ordered=True, fields will sorted by creation index.

    :param attrs: Mapping of class attributes
    :param type field_class: Base field class
    :param bool pop: Remove matching fields
    """
    fields = [
        (field_name, field_value)
        for field_name, field_value in attrs.items()
        if is_instance_or_subclass(field_value, field_class)
    ]
    if pop:  # pragma: no cover
        for field_name, _ in fields:
            del attrs[field_name]
    if ordered:
        fields.sort(key=lambda pair: pair[1]._creation_index)
    return fields


def _get_fields_by_mro(klass, field_class, ordered=False):
    """
    Collect fields from a class, following its method resolution order. The
    class itself is excluded from the search; only its parents are checked. Get
    fields from ``_declared_fields`` if available, else use ``__dict__``.

    :param type klass: Class whose fields to retrieve
    :param type field_class: Base field class
    """
    mro = inspect.getmro(klass)
    # Loop over mro in reverse to maintain correct order of fields
    return sum(
        (
            _get_fields(
                getattr(base, "_declared_fields", base.__dict__),
                field_class,
                ordered=ordered,
            )
            for base in mro[:0:-1]
        ),
        [],
    )


VALID_CHAR_SET = set(string.ascii_uppercase + string.digits + "_")


def _validate_field_name(field_name):
    """
    Validate config field name naming convention. Only [A-Z0-9_] is allowed.
    lowercase is not allowed.
    """
    if len(set(field_name).difference(VALID_CHAR_SET)):
        msg = "'{}' is not a valid field name, only [A-Z0-9_] is allowed!" \
            .format(field_name)
        raise ValueError(msg)


class ConfigMeta(type):
    """
    Config class meta class. Collect declared :class:`Field`, assign field name.
    """

    def __new__(cls, name, bases, attrs):
        cls_fields = _get_fields(attrs, Field, pop=False, ordered=True)
        klass = super(ConfigMeta, cls).__new__(cls, name, bases, attrs)
        inherited_fields = _get_fields_by_mro(klass, Field, ordered=True)

        # Assign _declared_fields on class
        klass._declared_fields = OrderedDict(inherited_fields + cls_fields)
        klass._constant_fields = OrderedDict([
            (name, field)
            for name, field in klass._declared_fields.items()
            if isinstance(field, Constant)
        ])
        klass._deriable_fields = OrderedDict([
            (name, field)
            for name, field in klass._declared_fields.items()
            if isinstance(field, Derivable)
        ])
        for name, field in klass._declared_fields.items():
            _validate_field_name(name)
            field.name = name
        return klass


class BaseConfigClass(object):
    """
    Config class base class.

    - :attr:`BaseConfigClass._declared_fields`:
    - :attr:`BaseConfigClass._constant_fields`:
    - :attr:`BaseConfigClass._deriable_fields`:
    """
    _declared_fields = OrderedDict()  # type: typing.Dict[str: Field]
    _constant_fields = OrderedDict()  # type: typing.Dict[str: Constant]
    _deriable_fields = OrderedDict()  # type: typing.Dict[str: Derivable]

    # --- constructor method
    def __init__(self, **kwargs):
        self.__pre_init_hook()
        for name, field in self._declared_fields.items():
            if name in kwargs:
                field.set_value(kwargs[name])

            field._config_object = self

    def __pre_init_hook(self):
        """
        All declared fields is a mutable :class:`Field` Instance, defined
        in Class level. So when you creating a new instance, the class
        level fields have to be deep copied to the config instance.
        """
        self._declared_fields = OrderedDict([
            (attr, copy.deepcopy(field))
            for attr, field in self._declared_fields.items()
        ])
        self._constant_fields = OrderedDict([
            (attr, self._declared_fields[attr])
            for attr, _ in self._constant_fields.items()
        ])
        self._deriable_fields = OrderedDict([
            (attr, self._declared_fields[attr])
            for attr, _ in self._deriable_fields.items()
        ])

        for attr, field in self._declared_fields.items():
            setattr(self, attr, field)

    @classmethod
    def from_dict(cls, dct):
        """
        A factory classmethod construct config object from dict.
        Only loads constant field. If a key is an undefined field,
        it automatically been ignored.

        :type dct: dict
        :rtype: BaseConfigClass
        """
        cfg = cls()
        for key, value in dct.items():
            if key in cfg._constant_fields:
                cfg._constant_fields[key].set_value(value)
        return cfg

    @classmethod
    def from_json_str(cls, json_str):
        """
        A factory classmethod construct config object from json string.
        json string can includes comments.
        Only loads constant field. If a key is an undefined field,
        it automatically been ignored.

        :type json_str: str
        :rtype: BaseConfigClass
        """
        return cls.from_dict(json_loads(json_str))

    @classmethod
    def from_json_file(cls, json_file):  # pragma: no cover
        """
        A factory classmethod construct config object from a json file.
        json string can includes comments.
        Only loads constant field. If a key is an undefined field,
        it automatically been ignored.

        :type json_file: str
        :rtype: BaseConfigClass
        """
        return cls.from_dict(json_load(json_file))

    @classmethod
    def from_env_var(cls, prefix):  # pragma: no cover
        """
        A factory classmethod construct config object from environment variables.
        Only loads constant field. If a key is an undefined field,
        it automatically been ignored.

        :type json_file: str
        :rtype: BaseConfigClass
        """
        cfg = cls()
        cfg.update_from_env_var(prefix=prefix)
        return cfg

    def update(self, dct):
        """
        Update constant config values from a dictionary.
        Only those fields defines as Constant value will be loaded.
        If a key is an undefined field, it automatically been ignored.

        :type dct: dict
        :rtype: dict
        :return: loaded data
        """
        loaded_data = dict()
        for key, value in dct.items():
            if key in self._constant_fields:
                self._constant_fields[key].set_value(value)
                loaded_data[key] = value
        return loaded_data

    def update_from_raw_json_file(self):
        """
        Update constant config values from the :attr:`BaseConfigClass.CONFIG_RAW_JSON_FILE`.

        :rtype: dict
        :return: loaded data
        """
        dct = json_load(self.CONFIG_RAW_JSON_FILE)
        return self.update(dct)

    def update_from_env_var(self, prefix):
        """
        Update constant config values from environment variables.

        :type prefix: str
        :param prefix: a prefix used in all related environment variable.

        :rtype: dict
        :return: loaded data
        """
        dct = {
            key.replace(prefix, "", 1): value
            for key, value in os.environ.items()
            if key.replace(prefix, "", 1)
        }
        self.update(dct)
        return dct

    def to_dict(self,
                check_dont_dump=True,
                check_printable=False,
                ignore_na=False,
                prefix=""):
        """
        Dump config values to dictionary.

        :type check_dont_dump: bool
        :param check_dont_dump: if True, then it will check if a field has
            a True value ``dont_dump`` flag, then :class:`DontDumpError` error
            is raised.

        :type check_printable: bool
        :param check_printable: if True, then it will check if a field has
            a False value ``printable`` flag, then it returns **HIDDEN**.

        :type ignore_na: bool
        :param ignore_na: if True, then :class:`ValueNotSetError` error will be
            ignored.

        :type prefix: str
        :param prefix: a prefix that appended to the left of every field

        :rtype: dict
        """
        dct = OrderedDict()
        for attr, value in self._declared_fields.items():
            key = prefix + attr
            try:
                dct[key] = value.get_value(
                    check_dont_dump=check_dont_dump,
                    check_printable=check_printable,
                )
            except DontDumpError:
                pass
            except ValueNotSetError as e:
                if ignore_na:
                    pass
                else:
                    raise e
            except Exception as e:
                raise e
        return dct

    def to_json(self,
                check_dont_dump=True,
                check_printable=False,
                ignore_na=False,
                prefix=""):
        """
        Dump config values to json.

        :type check_dont_dump: bool
        :param check_dont_dump: if True, then it will check if a field has
            a True value ``dont_dump`` flag, then :class:`DontDumpError` error
            is raised.

        :type check_printable: bool
        :param check_printable: if True, then it will check if a field has
            a False value ``printable`` flag, then it returns **HIDDEN**.

        :type ignore_na: bool
        :param ignore_na: if True, then :class:`ValueNotSetError` error will be
            ignored.

        :type prefix: str
        :param prefix: a prefix that appended to the left of every field

        :rtype: str
        """
        return json.dumps(
            self.to_dict(
                check_dont_dump=check_dont_dump,
                check_printable=check_printable,
                ignore_na=ignore_na,
                prefix=prefix,
            ),
            indent=4, sort_keys=False,
        )

    def __repr__(self):
        return "Config({})".format(
            self.to_json(check_dont_dump=False, check_printable=True)
        )

    def pprint(self):
        print(self.__repr__())

    def validate(self):
        for field in self._declared_fields.values():
            field.validate()

    # --- Runtime Detection ---
    @classmethod
    def is_aws_ec2_amz_linux_runtime(cls):  # pragma: no cover
        """
        Check whether it is Amazon Linux EC2 runtime.

        :rtype: bool
        """
        if os.environ["HOME"].endswith("ec2-user"):
            return True
        else:
            return False

    @classmethod
    def is_aws_ec2_redhat_runtime(cls):  # pragma: no cover
        """
        Check whether it is RedHat AWS EC2 runtime.

        :rtype: bool
        """
        if os.environ["HOME"].endswith("ec2-user"):
            return True
        else:
            return False

    @classmethod
    def is_aws_ec2_freebsd_runtime(cls):  # pragma: no cover
        """
        Check whether it is FreeBSD AWS EC2 runtime.

        :rtype: bool
        """
        if os.environ["HOME"].endswith("ec2-user"):
            return True
        else:
            return False

    @classmethod
    def is_aws_lambda_runtime(cls):  # pragma: no cover
        """
        Check whether it is Amazon Lambda Function runtime.

        Ref: https://docs.aws.amazon.com/lambda/latest/dg/lambda-environment-variables.html

        :rtype: bool
        """
        if "AWS_LAMBDA_FUNCTION_NAME" in os.environ:
            return True
        else:
            return False

    @classmethod
    def is_aws_code_build_runtime(cls):  # pragma: no cover
        """
        Check whether it is AWS Code Build runtime.

        Ref: https://docs.aws.amazon.com/codebuild/latest/userguide/build-env-ref-env-vars.html

        :rtype: bool
        """
        if "CODEBUILD_BUILD_ID" in os.environ:
            return True
        else:
            return False

    # CI
    @classmethod
    def is_ci_runtime(cls):  # pragma: no cover
        """
        Check whether it is CI runtime.

        :rtype: bool
        """
        if "CI" in os.environ:
            if os.environ["CI"]:
                return True
            else:
                return False
        else:
            return False

    @classmethod
    def is_circle_ci_runtime(cls):  # pragma: no cover
        """
        Check whether it is CircleCI runtime.

        Ref: https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables

        :rtype: bool
        """
        if "CIRCLECI" in os.environ:
            if os.environ["CIRCLECI"]:
                return True
            else:
                return False
        else:
            return False

    @classmethod
    def is_travis_ci_runtime(cls):  # pragma: no cover
        """
        Check whether it is TravisCI runtime.

        Ref: https://docs.travis-ci.com/user/environment-variables/#default-environment-variables

        :rtype: bool
        """
        if "TRAVIS" in os.environ:
            if os.environ["TRAVIS"]:
                return True
            else:
                return False
        else:
            return False

    @classmethod
    def is_gitlab_ci_runtime(cls):  # pragma: no cover
        """
        Check whether it is GitlabCI runtime.

        Ref: https://docs.gitlab.com/ee/ci/variables/

        :rtype: bool
        """
        if "GITLAB_CI" in os.environ:
            if os.environ["GITLAB_CI"]:
                return True
            else:
                return False
        else:
            return False

    # --- config file path management
    CONFIG_DIR = NOTHING  # type: str

    def _join_config_dir(self, filename):
        if self.CONFIG_DIR is NOTHING:
            raise ValueError("You have to specify `{}.CONFIG_DIR`!".format(
                self.__class__.__name__))
        if not os.path.exists(self.CONFIG_DIR):
            raise ValueError("`{}.CONFIG_DIR` ('{}') doesn't exist!".format(
                self.__class__.__name__, self.CONFIG_DIR))
        return os.path.join(self.CONFIG_DIR, filename)

    @property
    def CONFIG_RAW_JSON_FILE(self):
        return self._join_config_dir("config-raw.json")

    @property
    def CONFIG_FINAL_JSON_FILE_FOR_PYTHON(self):
        return self._join_config_dir("config-final-for-python.json")

    @property
    def CONFIG_FINAL_JSON_FILE_FOR_SHELL_SCRIPT(self):
        return self._join_config_dir("config-final-for-shell-script.json")

    @property
    def CONFIG_FINAL_JSON_FILE_FOR_CLOUDFORMATION(self):
        return self._join_config_dir("config-final-for-cloudformation.json")

    @property
    def CONFIG_FINAL_JSON_FILE_FOR_SAM(self):
        return self._join_config_dir("config-final-for-sam.json")

    @property
    def CONFIG_FINAL_JSON_FILE_FOR_SERVERLESS(self):
        return self._join_config_dir("config-final-for-serverless.json")

    @property
    def CONFIG_FINAL_JSON_FILE_FOR_TERRAFORM(self):
        return self._join_config_dir("config-final-for-terraform.json")

    # --- Custom logic for different devops tools
    def to_python_json_config_data(self):
        return self.to_dict()

    def to_shell_script_config_data(self):
        return self.to_dict()

    def to_cloudformation_config_data(self):
        def to_big_camel_case(text):
            return "".join([
                word[0].upper() + word[1:].lower()
                for word in text.split("_")
            ])

        return OrderedDict([
            (to_big_camel_case(key), value)
            for key, value in self.to_dict().items()
        ])

    def to_sam_config_data(self):
        return self.to_dict()

    def to_serverless_config_data(self):
        return self.to_dict()

    def to_terraform_config_data(self):
        return self.to_dict()

    def _dump_for_xxx_config_file(self,
                                  to_config_data_meth,
                                  config_json_file_path):
        json_str = json_dumps(to_config_data_meth())
        write_text(json_str, config_json_file_path)

    def dump_python_json_config_file(self):
        self._dump_for_xxx_config_file(
            self.to_python_json_config_data,
            self.CONFIG_FINAL_JSON_FILE_FOR_PYTHON,
        )

    def dump_shell_script_json_config_file(self):
        self._dump_for_xxx_config_file(
            self.to_shell_script_config_data,
            self.CONFIG_FINAL_JSON_FILE_FOR_SHELL_SCRIPT,
        )

    def dump_cloudformation_json_config_file(self):
        self._dump_for_xxx_config_file(
            self.to_cloudformation_config_data,
            self.CONFIG_FINAL_JSON_FILE_FOR_CLOUDFORMATION,
        )

    def dump_sam_json_config_file(self):
        self._dump_for_xxx_config_file(
            self.to_sam_config_data,
            self.CONFIG_FINAL_JSON_FILE_FOR_SAM,
        )

    def dump_serverless_json_config_file(self):
        self._dump_for_xxx_config_file(
            self.to_serverless_config_data,
            self.CONFIG_FINAL_JSON_FILE_FOR_SERVERLESS,
        )

    def dump_terraform_json_config_file(self):
        self._dump_for_xxx_config_file(
            self.to_terraform_config_data,
            self.CONFIG_FINAL_JSON_FILE_FOR_TERRAFORM,
        )


@add_metaclass(ConfigMeta)
class ConfigClass(BaseConfigClass):
    pass


# --- Command Line Interface ---

class SubCommands:
    read_json_value = "read-json-value"
    get_config_value = "get-config-value"
    import_config_value = "import-config-value"


def read_json_value(path, field):
    """
    Return a value of a field from a Json file, ignoring the comments
    """
    # find absolute path
    cwd = os.getcwd()

    if not os.path.isabs(path):
        file_path = os.path.abspath(os.path.join(cwd, path))
    else:
        file_path = path

    # fix json_path
    json_path = field
    if json_path.startswith("$."):
        json_path = json_path.replace("$.", "", 1)

    # read data
    with open(file_path, "rb") as f:
        data = json.loads(strip_comments(f.read().decode("utf-8")))

    # access value
    value = data
    for part in json_path.split("."):
        if part in value:
            value = value[part]
        else:
            raise ValueError(
                "'$.{}' not found in {}".format(json_path, file_path))
    return value


def get_config_value(module, field):
    """
    Initialize a Config Class defined in a python module, and get the value
    of a field. The module has to be able to be import.
    """
    chunks = module.split(".")
    module_object = importlib.import_module(".".join(chunks[:-1]))
    klass = getattr(module_object, chunks[-1])
    cfg_obj = klass()
    attribute = getattr(cfg_obj, field)
    if isinstance(attribute, Field):
        return attribute.get_value()
    elif callable(attribute):
        return attribute()
    else:
        return attribute


def import_config_value(sys_path, module, field):
    """
    Initialize a Config Class defined in a python module, and get the value
    of a field. The module has to be able to be import when arg ``sys_path``
    been add to sys.path.
    """
    if sys_path not in sys.path:
        sys.path.append(sys_path)
    return get_config_value(module, field)


parser = argparse.ArgumentParser(
    prog="configirl",
)

subparser = parser.add_subparsers(
    title="sub commands",
    dest="sub_command",
)

read_json_value_parser = subparser.add_parser(
    SubCommands.read_json_value,
    description=(
        "read config value from a json file."
    ),
)
read_json_value_parser.add_argument(
    "--path",
    type=str,
    metavar="path",
    nargs=1,
    help="The json file path",
    required=True,
)
read_json_value_parser.add_argument(
    "--field",
    type=str,
    metavar="field_name",
    nargs=1,
    help="The value of json field you want to access. For example: STAGE",
    required=True,
)

get_config_value_parser = subparser.add_parser(
    SubCommands.get_config_value,
    description=(
        "get config value by importing a Config object from a python module, "
        "and initializing a config object, "
        "and then call ``Config().FIELD_NAME.get_value()`` method."
    ),
)
get_config_value_parser.add_argument(
    "--module",
    type=str,
    metavar="module_name",
    nargs=1,
    help=(
        "The path you want to add to ``sys.path`` that allows python to find your module. "
        "For example: path-to/configirl-project"
    ),
    required=True,
)
get_config_value_parser.add_argument(
    "--field",
    type=str,
    metavar="field_name",
    nargs=1,
    help=(
        "The value of config field you want to access. "
        "For example: ENVIRONMENT_NAME"
    ),
    required=True,
)

import_config_value_parser = subparser.add_parser(
    SubCommands.import_config_value,
    description=(
        "import config value by importing a Config object from a python module, "
        "where locate at custom system path, "
        "and initializing a config object, "
        "and then call ``Config().FIELD_NAME.get_value()`` method."
    ),
)
import_config_value_parser.add_argument(
    "--sys_path",
    type=str,
    metavar="sys_path",
    nargs=1,
    help=(
        "The path you want to add to ``sys.path`` that allows python to find your module. "
        "For example: path-to/configirl-project"
    ),
    required=True,
)
import_config_value_parser.add_argument(
    "--module",
    type=str,
    metavar="module_name",
    nargs=1,
    help=(
        "The config class module name. "
        "For example: configirl.tests.config.Config"
    ),
    required=True,
)
import_config_value_parser.add_argument(
    "--field",
    type=str,
    metavar="field_name",
    nargs=1,
    help=(
        "The value of config field you want to access. "
        "For example: ENVIRONMENT_NAME"
    ),
    required=True,
)


def main():  # pragma: no cover
    """
    Command Line Interface entry point.
    """
    args = parser.parse_args()
    if args.sub_command == SubCommands.read_json_value:
        print(read_json_value(path=args.path[0], field=args.field[0]))
    elif args.sub_command == SubCommands.get_config_value:
        print(get_config_value(module=args.module[0], field=args.field[0]))
    elif args.sub_command == SubCommands.import_config_value:
        print(import_config_value(
            sys_path=args.sys_path[0], module=args.module[0], field=args.field[0]))
    else:
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""
This module detects the current operation system.
"""

from __future__ import unicode_literals
import platform


class OSEnum(object):
    """
    Operation System name enumeration.
    """
    windows = "Windows"
    macOS = "Darwin"
    linux = "Linux"
    java = "Java"
    unknown = ""


OS_NAME = platform.system()  # type: str
"""
Current OS name
"""

IS_WINDOWS = OS_NAME == OSEnum.windows  # type: bool
"""
Boolean value that indicate if the current OS is Windows 
"""

IS_MACOS = OS_NAME == OSEnum.macOS  # type: bool
"""
Boolean value that indicate if the current OS is MacOS 
"""

IS_LINUX = OS_NAME == OSEnum.linux  # type: bool
"""
Boolean value that indicate if the current OS is Linux 
"""

IS_JAVA = OS_NAME == OSEnum.java  # type: bool
"""
Boolean value that indicate if the current OS is Java 
"""

if OS_NAME not in (OSEnum.windows, OSEnum.macOS, OSEnum.linux):
    raise EnvironmentError("Not supported OS: {}".format(OS_NAME))

OPEN_COMMAND = None
"""
The OS dependent command that open a file in default application.
"""
if OS_NAME == OSEnum.windows:
    OPEN_COMMAND = "start"
elif OS_NAME in (OSEnum.macOS, OSEnum.linux):
    OPEN_COMMAND = "open"
else:
    OPEN_COMMAND = "unknown"



# -*- coding: utf-8 -*-
import os

try:
    import typing
except:
    pass
import argparse
from pygitrepo.actions import Actions, actions
from pygitrepo.repo_config import RepoConfig
from pygitrepo.pkg.mini_colorma import Fore, Style

SUB_COMMAND = "sub_command"


def colorful(message):
    return Fore.CYAN + message + Style.RESET_ALL


parser = argparse.ArgumentParser(
    prog="pygitrepo",
)
subparser = parser.add_subparsers(
    title="sub commands",
    description="run automation script",
    prog="hello",
    dest=SUB_COMMAND,
)

# A dictionary object provide access to the underlying function using
# subcommand name
action_mapper = dict()  # type: typing.Dict[str, callable]
for method_name, method in Actions.__dict__.items():
    if not method_name.startswith("_"):
        subcommand_parser = subparser.add_parser(
            name=method._subcommand_name,
            help=Fore.CYAN + method._subcommand_help + Style.RESET_ALL
        )
        subcommand_parser.add_argument(
            "--do-dry-run",
            action="store_true",
            help="display info, doesn't take effect",
        )
        action_mapper[method._subcommand_name] = getattr(actions, method_name)


# manually add more important subc  ```````ommand parser
class AddtionalSubCommandEnum:
    get_value = None


for k in AddtionalSubCommandEnum.__dict__:
    if not k.startswith("_"):
        setattr(AddtionalSubCommandEnum, k, k.replace("_", "-"))

get_value_parser = subparser.add_parser(
    name=AddtionalSubCommandEnum.get_value,
    help=Fore.CYAN + "pass python data to shell scripts" + Style.RESET_ALL
)

get_value_parser.add_argument(dest="attr_name", type=str)


def get_value(config, attr_name):
    """
    echo the value of an attribute
    
    .. code-block:: bash
    
        dir_project_root="$(pgr get-value dir_project_root)"
    
    :type config: RepoConfig
    """
    if isinstance(getattr(RepoConfig, attr_name), property):
        print(getattr(config, attr_name))
    else:
        print(getattr(config, attr_name).get_value())


def main():  # pragma: no cover
    """
    Command Line Interface entry point.
    """
    # defined arguments stored in args, a named tuple object
    # additional undefined arguments stored in unknown
    args, unknown = parser.parse_known_args()
    print(args)
    if args.sub_command is None:
        parser.parse_args(["-h"])
        return

    repo_config = RepoConfig()
    repo_config.read_pygitrepo_config_file()

    if args.sub_command in action_mapper:
        action_mapper[args.sub_command](
            repo_config,
            _dry_run=args.do_dry_run,
            _args=unknown,
        )
    elif args.sub_command == AddtionalSubCommandEnum.get_value:
        get_value(repo_config, args.attr_name)
    else:
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""
Utility module for colored text print in terminal.
"""

import os

from .pkg.mini_colorma import Fore, Style

TAB = " " * 2


def pgr_print(message):
    """
    Syntax sugar for:

    [pygitrepo] ... message ... {Style.RESET_ALL}

    :type message: str
    """
    print("[pygitrepo] {}{}".format(message, Style.RESET_ALL))


def pgr_print_done(indent=0):
    """
    Syntax sugar for:

    [pygitrepo] {number_of_tab_indent}done

    :type indent: int
    """
    print("{reset}[pygitrepo] {cyan}{indent}done{reset}".format(
        cyan=Fore.CYAN,
        reset=Style.RESET_ALL,
        indent=indent * TAB,
    ))


def colorful_path(path):
    """
    :type path: str
    :rtype: str
    """
    if os.path.exists(path):
        color = Fore.GREEN
    else:
        color = Fore.RED
    return "{color}{path}{reset}".format(
        color=color,
        path=path,
        reset=Style.RESET_ALL,
    )


def print_path(title, path):
    """
    Print a description of a path and the absolute path. If path exists,
    path is in green, if not exists, path is in red.

    Example: "- virtual environment directory: /path-to/my_venv"

    :type title: str
    :type path: str
    """
    if os.path.exists(path):
        color = Fore.GREEN
    else:
        color = Fore.RED
    msg = "{reset}- {cyan}{title}: {colorful_path}".format(
        reset=Style.RESET_ALL,
        cyan=Fore.CYAN,
        color=color,
        title=title,
        colorful_path=colorful_path(path),
    )
    print(msg)


def print_line(msg):
    """
    Color print a message reset style before and after.

    :type msg: str
    :rtype: str
    """
    print("{reset}{msg}{reset}".format(reset=Style.RESET_ALL, msg=msg))



# -*- coding: utf-8 -*-

try:
    import typing
except:  # pragma: no cover
    pass

import os
import shutil
from re import findall


def split_s3_uri(s3_uri):
    """
    Split AWS S3 URI, returns bucket and key.

    :type s3_uri: str
    :rtype: typing.Tuple[str, str]
    """
    parts = s3_uri.split("/")
    bucket = parts[2]
    key = "/".join(parts[3:])
    return bucket, key


def join_s3_uri(bucket, key):
    """
    Join AWS S3 URI from bucket and key.

    :type bucket: str
    :type key: str
    :rtype: str
    """
    return "s3://{}/{}".format(bucket, key)


def s3_key_smart_join(parts, is_dir):
    """
    Note, it assume that there's no such double slack in your path. It ensure
    that there's only one consecutive "/" in the s3 key.

    :type parts: typing.List[str]
    :param parts: list of s3 key path parts, could have "/"

    :type is_dir: bool
    :param is_dir: if True, the s3 key ends with "/". otherwise enforce no
        tailing "/".

    :rtype: str

    Example::

        >>> s3_key_smart_join(parts=["/a/", "b/", "/c"], is_dir=True)
        a/b/c/

        >>> s3_key_smart_join(parts=["/a/", "b/", "/c"], is_dir=False)
        a/b/c
    """
    new_parts = list()
    for part in parts:
        new_parts.extend([chunk for chunk in part.split("/") if chunk])
    key = "/".join(new_parts)
    if is_dir:
        return key + "/"
    else:
        return key


def make_s3_console_url(bucket=None, prefix=None, s3_uri=None):
    """
    Return an AWS Console url that you can use to open it in your browser.

    :type bucket: str
    :type prefix: str
    :type s3_uri: str
    :rtype: str
    """
    if s3_uri is None:
        if (bucket is None) or (prefix is None):
            raise ValueError
        else:
            pass
    else:
        if (bucket is not None) or (prefix is not None):
            raise ValueError
        bucket, prefix = split_s3_uri(s3_uri)

    if prefix.endswith("/"):
        s3_type = "buckets"
    else:
        s3_type = "object"
    return "https://s3.console.aws.amazon.com/s3/{s3_type}/{bucket}?prefix={prefix}".format(
        s3_type=s3_type,
        bucket=bucket,
        prefix=prefix
    )


def s3_uri_to_url(s3_uri):
    """
    Convert a S3 URI to AWS S3 Console url for preview.

    :type s3_uri
    :rtype: str
    """
    bucket, key = split_s3_uri(s3_uri)
    return make_s3_console_url(bucket=bucket, prefix=key)


def ensure_s3_object(s3_key_or_uri):
    """
    Raise exception if the string is not in valid format for a AWS S3 object
    """
    if s3_key_or_uri.endswith("/"):
        raise ValueError("'{}' doesn't represent s3 object!".format(s3_key_or_uri))


def ensure_s3_dir(s3_key_or_uri):
    """
    Raise exception if the string is not in valid format for a AWS S3 directory
    """
    if not s3_key_or_uri.endswith("/"):
        raise ValueError("'{}' doesn't represent s3 dir!".format(s3_key_or_uri))


def strip_comment_line_with_symbol(line, start):  # pragma: no cover
    """
    Strip comments from line string.
    """
    parts = line.split(start)
    counts = [len(findall(r'(?:^|[^"\\]|(?:\\\\|\\")+)(")', part))
              for part in parts]
    total = 0
    for nr, count in enumerate(counts):
        total += count
        if total % 2 == 0:
            return start.join(parts[:nr + 1]).rstrip()
    else:  # pragma: no cover
        return line.rstrip()


def strip_comments(string, comment_symbols=frozenset(('#', '//'))):  # pragma: no cover
    """
    Strip comments from json string.

    :param string: A string containing json with comments started by comment_symbols.
    :param comment_symbols: Iterable of symbols that start a line comment (default # or //).
    :return: The string with the comments removed.
    """
    lines = string.splitlines()
    for k in range(len(lines)):
        for symbol in comment_symbols:
            lines[k] = strip_comment_line_with_symbol(lines[k], start=symbol)
    return '\n'.join(lines)


def remove_if_exists(abspath):
    """
    Remove a file or a directory (and it's files) if exists.

    :type abspath: str
    """
    if not os.path.exists(abspath):
        return
    if os.path.isdir(abspath):
        shutil.rmtree(abspath)
    else:
        os.remove(abspath)


def makedir_if_not_exists(abspath):
    """
    Make a directory and all required parent folder if not exists.

    :type abspath: str
    """
    if os.path.exists(abspath):
        return
    os.makedirs(abspath)


def copy_python_code(from_dir, to_dir):
    """
    Copy all python source code from one directory to another. Skip
    ``__pycache__``, ``.pyc`` and ``.pyo`` files.

    :type from_dir: str
    :type to_dir: str
    """
    remove_if_exists(to_dir)
    pycache = "__pycache__"
    for dirname, _, basename_list in os.walk(from_dir):
        relpath = os.path.relpath(dirname, from_dir)
        target_dir = os.path.abspath(os.path.join(to_dir, relpath))
        if target_dir.endswith(pycache):
            continue
        else:
            makedir_if_not_exists(target_dir)
        for basename in basename_list:
            # ignore .pyc, the compiled byte code
            # and .pyo, the optimized import cache
            if basename.endswith(".pyc") or basename.endswith(".pyo"):
                continue
            else:
                source_path = os.path.join(dirname, basename)
                target_path = os.path.join(target_dir, basename)
                shutil.copyfile(source_path, target_path)



# -*- coding: utf-8 -*-

import os
import sys
import json
from .pkg.configirl import ConfigClass, Constant, Derivable
from . import constants
from .helpers import (
    join_s3_uri,
    s3_key_smart_join,
    strip_comments,
)
from .operation_system import (
    IS_WINDOWS, IS_MACOS, IS_LINUX,
)


class _BaseConfig(ConfigClass):
    """
    The abstract config class.
    """

    def ensure_attr_not_none(self, attr_name):
        """
        raise error if an config attribute value is None.
        """
        value = getattr(self, attr_name).get_value()
        if value is None:
            raise ValueError(
                "Please give '{}' a valid value other than 'None'!".format(
                    attr_name
                )
            )


class _RepoConfig(_BaseConfig):
    """
    Basic Python GitHub Repo related config.
    """
    PACKAGE_NAME = Constant(default=None)
    GITHUB_ACCOUNT = Constant(default=None)
    GITHUB_REPO_NAME = Constant(default=None)

    # The main python version you use for development
    DEV_PY_VER_MAJOR = Constant(default=None)
    DEV_PY_VER_MINOR = Constant(default=None)
    DEV_PY_VER_MICRO = Constant(default=None)

    # python versions you want to test in matrix test
    TOX_TEST_VERSIONS = Constant(default=lambda: list())

    @property
    def package_name(self):
        return self.PACKAGE_NAME.get_value()

    @property
    def package_name_slugify(self):
        """
        The slugified version of package name. No underscore, only use dash.
        """
        return self.PACKAGE_NAME.get_value().replace("_", "-")


class _DocConfig(_BaseConfig):
    """
    Documentation build related config.
    """
    DOC_HOST_RTD_PROJECT_NAME = Constant(default=None)
    """
    If you choose to host your document on https://readthedocs.org/, it is  
    the project name for https://readthedocs.org/, your documentation website
    url will become https://${DOC_HOST_RTD_PROJECT_NAME}.readthedocs.io/
    """

    DOC_HOST_AWS_PROFILE = Constant(default=None)
    """
    If you choose to host your document on AWS S3 as static site, it is the 
    AWS named profile you use to authenticate for uploading doc files to AWS S3 
    """

    DOC_HOST_S3_BUCKET = Constant(default=None)
    """
    If you choose to host your document on AWS S3 as static site, it is the 
    bucket you use to host this project.
    """


class _AWSLambdaConfig(_BaseConfig):
    AWS_LAMBDA_DEPLOY_AWS_PROFILE = Constant(default=None)
    """
    The AWS named profile you use to authenticate for AWS Lambda deployment.
    """

    AWS_LAMBDA_DEPLOY_S3_BUCKET = Constant(default=None)
    """
    The AWS S3 bucket for Lambda deployment
    """

    AWS_LAMBDA_BUILD_DOCKER_IMAGE = Constant(default=None)
    AWS_LAMBDA_BUILD_DOCKER_IMAGE_WORKSPACE_DIR = Constant(default=None)
    AWS_LAMBDA_TEST_DOCKER_IMAGE = Constant(default=None)

    def ensure_aws_lambda_deploy_s3_bucket(self):
        self.ensure_attr_not_none(self.AWS_LAMBDA_DEPLOY_S3_BUCKET.name)


class RepoConfig(
    _RepoConfig,
    _DocConfig,
    _AWSLambdaConfig,
):
    """
    Python Github Repo Config Object
    """
    DIR_CWD = Derivable(cache=True)
    """
    By default, it is where you originally call this script.

    let's say you are at ``/here`` and you have ``a.sh`` and ``b.py``::

        /
        /first/a.sh
        /first/second/b.py

    If you made a ``bash ./first/a.sh`` and ``a.sh`` actually calls
    ``python /first/second/b.py``. You will see the ``os.getcwd()`` still
    returns the corret dir where you initially run the bash command in ``b.py``
    """

    @DIR_CWD.getter
    def get_DIR_CWD(self):
        return os.getcwd()

    def is_pygitrepo_root_dir(self, path):
        """
        Identify if a path is a valid python git repo directory.
        It should have a ``.git`` directory and a pygitrepo config json file.

        :type path: str
        :rtype: bool
        """
        dot_git_exists = os.path.exists(os.path.join(path, ".git"))
        config_exists = os.path.exists(os.path.join(path, constants.PYGITREPO_CONFIG_FILE))
        return dot_git_exists and config_exists

    def locate_pygitrepo_root_dir(self):
        """
        Try to locate the pygitrepo root directory. It scan backward from the
        current directory to the "/" (MacOS or Linux) or "C://" (Windows).
        The first directory meets the criterial defined in :
        """
        cwd = self.DIR_CWD.get_value()
        for _ in range(1000):  # a path should have less than 1000 parts
            if self.is_pygitrepo_root_dir(cwd):
                return cwd
            else:
                previous_cwd = cwd
                cwd = os.path.dirname(cwd)
                if cwd == previous_cwd:
                    break
        raise EnvironmentError(
            "cannot locate a valid pygitrepo directory "
            "that has a `.git` dir and `{}` file".format(constants.PYGITREPO_CONFIG_FILE)
        )

    def read_pygitrepo_config_file(self):
        """
        Try to locate the pygitrepo config file.
        """
        pygitrepo_root_dir = self.locate_pygitrepo_root_dir()
        p = os.path.join(
            pygitrepo_root_dir,
            constants.PYGITREPO_CONFIG_FILE
        )
        with open(p, "rb") as f:
            json_content = f.read().decode("utf-8")
            config_data = json.loads(strip_comments(json_content))
        self.update(config_data)

    DIR_HOME = Derivable(cache=True)
    r"""
    The current User home directory.
    
    - On Windows: C:\Users\${username}
    - On Mac: /Users/${username}
    - On Linux: depends on what Linux. on Ubuntu / Redhat / CentOS 
        it is /home/${username}
    """

    @DIR_HOME.getter
    def get_DIR_HOME(self):
        return os.path.expanduser("~")

    # === Code File Structure
    # --- python project basics
    DIR_PROJECT_ROOT = Derivable(cache=True)

    @DIR_PROJECT_ROOT.getter
    def get_DIR_PROJECT_ROOT(self):
        return self.locate_pygitrepo_root_dir()

    @property
    def dir_project_root(self):
        return self.DIR_PROJECT_ROOT.get_value()

    @property
    def path_readme(self):
        return os.path.join(self.dir_project_root, "README.rst")

    @property
    def dir_python_lib(self):
        return os.path.join(self.dir_project_root, self.PACKAGE_NAME.get_value())

    @property
    def path_version_file(self):
        return os.path.join(self.dir_python_lib, "_version.py")

    @property
    def package_version(self):
        sys.path.append(self.dir_python_lib)
        try:
            from _version import __version__
            return __version__
        except:
            return constants.UNKNOWN

    @property
    def path_requirements_file(self):
        return os.path.join(self.dir_project_root, "requirements.txt")

    @property
    def path_requirements_dev_file(self):
        return os.path.join(self.dir_project_root, "requirements-dev.txt")

    @property
    def path_requirements_doc_file(self):
        return os.path.join(self.dir_project_root, "requirements-doc.txt")

    @property
    def path_requirements_test_file(self):
        return os.path.join(self.dir_project_root, "requirements-test.txt")

    @property
    def dir_pypi_build(self):
        return os.path.join(self.dir_project_root, "build")

    @property
    def dir_pypi_distribute(self):
        return os.path.join(self.dir_project_root, "dist")

    @property
    def dir_pypi_egg(self):
        return os.path.join(
            self.dir_project_root,
            "{}.egg-info".format(self.PACKAGE_NAME.get_value())
        )

    @property
    def path_fix_code_style_script(self):
        return os.path.join(self.dir_project_root, "fix_code_style.py")

    # --- testing
    @property
    def dir_tests(self):
        return os.path.join(self.dir_project_root, "tests")

    @property
    def dir_unit_tests(self):
        return self.dir_tests

    @property
    def dir_integration_tests(self):
        return os.path.join(self.dir_project_root, "tests_integration")

    @property
    def dir_pytest_cache(self):
        return os.path.join(self.dir_project_root, ".pytest_cache")

    @property
    def dir_codecov_yml(self):
        return os.path.join(self.dir_project_root, "codecov.yml")

    @property
    def path_coverage_config(self):
        return os.path.join(self.dir_project_root, ".coveragerc")

    @property
    def dir_coverage_annotate(self):
        return os.path.join(self.dir_project_root, ".coverage.annotate")

    @property
    def dir_pyenv_local_versions_for_tox(self):
        try:
            return " ".join(self.TOX_TEST_VERSIONS.get_value())
        except:
            return constants.UNKNOWN

    @property
    def dir_tox_cache(self):
        return os.path.join(self.dir_project_root, ".tox")

    # --- sphinx doc
    @property
    def dir_sphinx_doc(self):
        return os.path.join(self.dir_project_root, "docs")

    @property
    def dir_sphinx_doc_source(self):
        return os.path.join(self.dir_sphinx_doc, "source")

    @property
    def dir_sphinx_doc_source_conf_py(self):
        return os.path.join(self.dir_sphinx_doc_source, "conf.py")

    @property
    def dir_sphinx_doc_build(self):
        return os.path.join(self.dir_sphinx_doc, "build")

    @property
    def dir_sphinx_doc_build_html(self):
        return os.path.join(self.dir_sphinx_doc_build, "html")

    @property
    def path_sphinx_doc_build_html_index(self):
        return os.path.join(self.dir_sphinx_doc_build_html, "index.html")

    @property
    def path_readthedocs_yml(self):
        return os.path.join(self.dir_project_root, "readthedocs.yml")

    @property
    def url_rtd_doc(self):
        return "https://{}.readthedocs.io/".format(self.DOC_HOST_RTD_PROJECT_NAME.get_value())

    @property
    def url_s3_doc_latest(self):
        """
        The document website url for latest doc on s3.
        """
        return "https://{bucket}.s3.amazonaws.com/docs/{package_name}/latest/index.html".format(
            bucket=self.DOC_HOST_S3_BUCKET.get_value(),
            package_name=self.PACKAGE_NAME.get_value(),
        )

    @property
    def url_s3_doc_versioned(self):
        """
        The document website url for versioned doc on s3.
        """
        return "https://{bucket}.s3.amazonaws.com/docs/{package_name}/{version}/index.html".format(
            bucket=self.DOC_HOST_S3_BUCKET.get_value(),
            package_name=self.PACKAGE_NAME.get_value(),
            version=self.package_version,
        )

    @property
    def s3_uri_doc_dir_latest(self):
        return "s3://{bucket}/docs/{package_name}/latest".format(
            bucket=self.DOC_HOST_S3_BUCKET.get_value(),
            package_name=self.PACKAGE_NAME.get_value(),
        )

    @property
    def s3_uri_doc_dir_versioned(self):
        return "s3://{bucket}/docs/{package_name}/{version}/".format(
            bucket=self.DOC_HOST_S3_BUCKET.get_value(),
            package_name=self.PACKAGE_NAME.get_value(),
            version=self.package_version,
        )

    # === Pyenv
    @property
    def path_bin_global_python(self):
        if IS_WINDOWS:
            return "/c/Python{}.{}/python.exe".format(self.DEV_PY_VER_MAJOR, self.DEV_PY_VER_MINOR)
        elif IS_MACOS or IS_LINUX:
            return os.path.join(
                self.DIR_HOME.get_value(),
                ".pyenv",
                "shims",
                "python{}.{}".format(
                    self.DEV_PY_VER_MAJOR.get_value(),
                    self.DEV_PY_VER_MINOR.get_value(),
                )
            )
        else:
            raise EnvironmentError

    # === Virtualenv
    @property
    def venv_name(self):
        return "{}_venv".format(self.PACKAGE_NAME.get_value())

    @property
    def dir_all_python_versioned_venv(self):
        if IS_WINDOWS or IS_MACOS or IS_LINUX:
            return os.path.join(
                self.DIR_HOME.get_value(),
                "venvs",
                "python",
                "{}.{}.{}".format(
                    self.DEV_PY_VER_MAJOR.get_value(),
                    self.DEV_PY_VER_MINOR.get_value(),
                    self.DEV_PY_VER_MICRO.get_value(),
                ),
            )
        else:
            raise ValueError

    @property
    def dir_venv(self):
        """
        The virtual environment directory location. It should have
        a ``bin`` (MacOS) or a ``Script`` (Windows) folder in it.
        """
        return os.path.join(self.dir_all_python_versioned_venv, self.venv_name)

    @property
    def dir_venv_site_packages(self):
        if IS_WINDOWS:
            return os.path.join(self.dir_venv, "Lib", "site-packages")
        elif IS_MACOS or IS_LINUX:
            return os.path.join(
                self.dir_venv,
                "lib",
                "python{}.{}".format(
                    self.DEV_PY_VER_MAJOR.get_value(),
                    self.DEV_PY_VER_MINOR.get_value()
                ),
                "site-packages",
            )
        else:
            raise Exception

    @property
    def dir_venv_site_packages_64(self):
        if IS_WINDOWS:
            return os.path.join(self.dir_venv, "Lib64", "site-packages")
        elif IS_MACOS or IS_LINUX:
            return os.path.join(
                self.dir_venv,
                "lib64",
                "python{}.{}".format(
                    self.DEV_PY_VER_MAJOR.get_value(),
                    self.DEV_PY_VER_MINOR.get_value()),
                "site-packages",
            )
        else:
            raise Exception

    @property
    def dir_venv_site_packages_INSTALLED(self):
        return os.path.join(
            self.dir_venv_site_packages,
            self.PACKAGE_NAME.get_value()
        )

    @property
    def dir_venv_site_packages_egg_link(self):
        return os.path.join(
            self.dir_venv_site_packages,
            "{}.egg-link".format(self.package_name_slugify)
        )

    # --- venv/bin
    @property
    def dir_venv_bin(self):
        if IS_WINDOWS:
            return os.path.join(self.dir_venv, "Scripts")
        elif IS_MACOS or IS_LINUX:
            return os.path.join(self.dir_venv, "bin")
        else:
            raise Exception

    @property
    def path_venv_bin_python(self):
        return os.path.join(self.dir_venv_bin, "python")

    @property
    def path_venv_bin_pip(self):
        return os.path.join(self.dir_venv_bin, "pip")

    @property
    def path_venv_bin_activate(self):
        return os.path.join(self.dir_venv_bin, "activate")

    @property
    def path_venv_bin_pytest(self):
        return os.path.join(self.dir_venv_bin, "pytest")

    @property
    def path_venv_bin_sphinx_quickstart(self):
        return os.path.join(self.dir_venv_bin, "sphinx-quickstart")

    @property
    def path_venv_bin_twine(self):
        return os.path.join(self.dir_venv_bin, "twine")

    @property
    def path_venv_bin_tox(self):
        return os.path.join(self.dir_venv_bin, "tox")

    @property
    def path_venv_bin_jupyter(self):
        return os.path.join(self.dir_venv_bin, "jupyter")

    @property
    def path_venv_bin_ansible(self):
        return os.path.join(self.dir_venv_bin, "ansible")

    @property
    def path_venv_bin_aws(self):
        return os.path.join(self.dir_venv_bin, "aws")

    @property
    def path_venv_bin_chalice(self):
        return os.path.join(self.dir_venv_bin, "chalice")

    @property
    def path_venv_bin_flask(self):
        return os.path.join(self.dir_venv_bin, "flask")

    @property
    def path_venv_bin_configirl(self):
        return os.path.join(self.dir_venv_bin, "configirl")

    # === AWS CLI
    @property
    def aws_cli_profile_arg_doc_host(self):
        doc_host_aws_profile = self.DOC_HOST_AWS_PROFILE.get_value()
        if doc_host_aws_profile is None:
            return ""
        else:
            return doc_host_aws_profile

    @property
    def aws_cli_profile_arg_lambda_deploy(self):
        aws_lambda_deploy_aws_profile = self.AWS_LAMBDA_DEPLOY_AWS_PROFILE.get_value()
        if aws_lambda_deploy_aws_profile is None:
            return ""
        else:
            return aws_lambda_deploy_aws_profile

    # ==========================================================================
    # AWS Lambda
    # ==========================================================================
    # --- local ---
    @property
    def dir_lambda_build(self):
        return os.path.join(self.dir_pypi_build, "lambda")

    @property
    def path_lambda_build_source(self):
        return os.path.join(self.dir_lambda_build, "source.zip")

    @property
    def path_lambda_build_layer(self):
        return os.path.join(self.dir_lambda_build, "layer.zip")

    @property
    def path_lambda_build_deploy_package(self):
        return os.path.join(self.dir_lambda_build, "deploy-pkg.zip")

    # --- s3 ---
    @property
    def s3_key_lambda_deploy_dir(self):
        """
        example: lambda/my_package/
        """
        return s3_key_smart_join(
            parts=[
                "lambda",
                self.PACKAGE_NAME.get_value(),
            ],
            is_dir=True,
        )

    @property
    def s3_key_lambda_deploy_versioned_dir(self):
        """
        example: lambda/my_package/0.0.1/
        """
        return s3_key_smart_join(
            parts=[
                self.s3_key_lambda_deploy_dir,
                self.package_version,
            ],
            is_dir=True,
        )

    @property
    def s3_uri_lambda_deploy_versioned_dir(self):
        """
        example: s3://bucket/lambda/my_package/0.0.1/
        """
        self.ensure_aws_lambda_deploy_s3_bucket()
        return join_s3_uri(
            bucket=self.AWS_LAMBDA_DEPLOY_S3_BUCKET.get_value(),
            key=self.s3_key_lambda_deploy_versioned_dir,
        )

    # --- versioned source
    @property
    def s3_key_lambda_deploy_versioned_source_dir(self):
        """
        example: lambda/my_package/0.0.1/source/
        """
        return s3_key_smart_join(
            parts=[
                self.s3_key_lambda_deploy_versioned_dir,
                "source",
            ],
            is_dir=True,
        )

    @property
    def s3_uri_lambda_deploy_versioned_source_dir(self):
        """
        example: s3://bucket/lambda/my_package/0.0.1/source/
        """
        self.ensure_aws_lambda_deploy_s3_bucket()
        return join_s3_uri(
            bucket=self.AWS_LAMBDA_DEPLOY_S3_BUCKET.get_value(),
            key=self.s3_key_lambda_deploy_versioned_source_dir,
        )

    # --- versioned layer
    @property
    def s3_key_lambda_deploy_versioned_layer_dir(self):
        """
        example: lambda/my_package/0.0.1/layer/
        """
        return s3_key_smart_join(
            parts=[
                self.s3_key_lambda_deploy_versioned_dir,
                "layer",
            ],
            is_dir=True,
        )

    @property
    def s3_uri_lambda_deploy_versioned_layer_dir(self):
        """
        example: s3://bucket/lambda/my_package/0.0.1/layer/
        """
        self.ensure_aws_lambda_deploy_s3_bucket()
        return join_s3_uri(
            bucket=self.AWS_LAMBDA_DEPLOY_S3_BUCKET.get_value(),
            key=self.s3_key_lambda_deploy_versioned_layer_dir,
        )

    # --- versioned deploy package
    @property
    def s3_key_lambda_deploy_versioned_deploy_pkg_dir(self):
        """
        example: lambda/my_package/0.0.1/deploy-pkg/
        """
        return s3_key_smart_join(
            parts=[
                self.s3_key_lambda_deploy_versioned_dir,
                "deploy-pkg",
            ],
            is_dir=True,
        )

    @property
    def s3_uri_lambda_deploy_versioned_deploy_pkg_dir(self):
        """
        example: s3://bucket/lambda/my_package/0.0.1/deploy-pkg/
        """
        self.ensure_aws_lambda_deploy_s3_bucket()
        return join_s3_uri(
            bucket=self.AWS_LAMBDA_DEPLOY_S3_BUCKET.get_value(),
            key=self.s3_key_lambda_deploy_versioned_deploy_pkg_dir,
        )

    @property
    def aws_lambda_layer_name(self):
        return self.PACKAGE_NAME.get_value()

    @property
    def url_lambda_layer_console(self):
        return "https://console.aws.amazon.com/lambda/home?#/layers/{layer_name}".format(
            layer_name=self.aws_lambda_layer_name
        )

    # AWS Lambda
    # AWS Lambda with Chalice
    @property
    def dir_lambda_app(self):
        """
        example: ${dir_project_root}/lambda_app
        """
        return os.path.join(self.dir_project_root, "lambda_app")

    @property
    def dir_aws_chalice(self):
        """
        example: ${dir_project_root}/lambda_app/.chalice
        """
        return os.path.join(self.dir_lambda_app, ".chalice")

    @property
    def path_aws_chalice_config_json(self):
        """
        example: ${dir_project_root}/lambda_app/.chalice/config.json
        """
        return os.path.join(self.dir_aws_chalice, "config.json")

    @property
    def path_aws_chalice_app_py(self):
        """
        example: ${dir_project_root}/lambda_app/app.py
        """
        return os.path.join(self.dir_lambda_app, "app.py")

    @property
    def dir_aws_chalice_vendor(self):
        """
        example: ${dir_project_root}/lambda_app/vendor
        """
        return os.path.join(self.dir_aws_chalice, "vendor")

    @property
    def dir_aws_chalice_vendor_source(self):
        """
        example: ${dir_project_root}/lambda_app/vendor/${package_name}
        """
        return os.path.join(
            self.dir_aws_chalice_vendor,
            self.PACKAGE_NAME.get_value(),
        )

    @property
    def dir_aws_chalice_deployed(self):
        """
        example: ${dir_project_root}/lambda_app/.chalice/deployed
        """
        return os.path.join(self.dir_aws_chalice, "deployed")



# -*- coding: utf-8 -*-

"""
This module implements the ``pygitrepo`` Python development workflow
best practice logics.
"""

from __future__ import print_function, unicode_literals

try:
    import typing
except ImportError:  # pragma: no cover
    pass

import os
import shutil
import subprocess
import functools
from zipfile import ZipFile

from .pkg.mini_six import input
from .pkg.fingerprint import fingerprint
from .repo_config import RepoConfig
from .operation_system import (
    IS_WINDOWS, IS_MACOS, IS_LINUX,
    OPEN_COMMAND
)
from .helpers import (
    remove_if_exists, makedir_if_not_exists,
    join_s3_uri,
    split_s3_uri,
    make_s3_console_url,
    s3_uri_to_url,
    s3_key_smart_join,
    ensure_s3_dir,
    copy_python_code,
)
from .color_print import (
    Fore, Style, TAB,
    pgr_print, pgr_print_done, print_path, print_line, colorful_path,
)


def subcommand(
    name=None,
    help=None,
):
    """
    A decorator that mark a function / class method a sub command for CLI.

    :type name: str
    :type help: str
    """

    def real_deco(func):
        if name is None:
            func._subcommand_name = func.__name__.replace("_", "-")
        else:
            func._subcommand_name = name.replace("_", "-")
        func._subcommand_help = help
        func._is_subcommand = True

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            res = func(*args, **kwargs)
            return res

        return wrapper

    return real_deco


class Actions(object):
    """
    A class container that includes ``pgr`` CLI interface logic.
    Each command method wrapped with a :func:`subcommand` decorator is a
    underlying logic for ``pgr ${subcommand}``.

    Usually each function has a ``**kwargs`` optional keyword arguments.
    It can be used to store optional command line arguments.
    """

    # ==============================================================================
    # Python Package Development Related
    # ==============================================================================
    @subcommand(
        help="** Create virtualenv for python package development.",
    )
    def venv_up(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}create a virtualenv at {reset}{dir_venv}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                dir_venv=config.dir_venv,
            )
        )
        if os.path.exists(config.dir_venv):
            pgr_print(
                "{cyan}{tab}skip, virtualenv directory already exists!".format(
                    cyan=Fore.CYAN,
                    tab=TAB,
                )
            )
        else:
            if _dry_run is False:
                # call the global virtualenv command
                subprocess.call([
                    "virtualenv",
                    "-p",
                    config.path_bin_global_python,
                    config.dir_venv,
                ])
                # call pip install command
                subprocess.call([
                    config.path_venv_bin_pip,
                    "install",
                    "--upgrade",
                    "pip",
                ])
            pgr_print_done(indent=1)

    @subcommand(
        help="** Remove the virtualenv for this project.",
    )
    def venv_remove(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}remove virtualenv at {reset}{dir_venv}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                dir_venv=config.dir_venv,
            )
        )
        if os.path.exists(config.dir_venv):
            if _dry_run is False:
                remove_if_exists(config.dir_venv)
            pgr_print_done(indent=1)
        else:
            pgr_print(
                "{cyan}{tab}skip, virtualenv directory doesn't exists!".format(
                    cyan=Fore.CYAN,
                    tab=TAB,
                )
            )

    @subcommand(
        help="** Clean temp files.",
    )
    def clean(
        self,
        config,
        ignore_tox=False,
        _dry_run=False,
        **kwargs
    ):
        """
        :type config: RepoConfig
        """
        pgr_print("{cyan}remove all temp files ...".format(cyan=Fore.CYAN))

        to_delete = [
            config.dir_pypi_build,
            config.dir_sphinx_doc_build,
            config.dir_coverage_annotate,
            config.dir_pytest_cache,
        ]

        if ignore_tox is False:
            to_delete.append(config.dir_tox_cache)

        for p in to_delete:
            pgr_print(
                "{cyan}  clean {reset}{path} {cyan}...".format(
                    cyan=Fore.CYAN,
                    reset=Style.RESET_ALL,
                    path=p,
                )
            )
            if _dry_run is False:
                remove_if_exists(p)

        pgr_print_done(indent=1)

    @subcommand(
        help="Uninstall the package from virtualenv.",
    )
    def pip_uninstall(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}uninstall '{package_name}' from virtualenv".format(
                cyan=Fore.CYAN,
                package_name=config.PACKAGE_NAME.get_value(),
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pip,
                "uninstall",
                "-y",
                config.PACKAGE_NAME.get_value(),
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="** Install the package to virtualenv in editable mode.",
    )
    def pip_dev_install(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self.pip_uninstall(config, _dry_run=_dry_run, **kwargs)
        pgr_print(
            "{cyan}install '{package_name}' to virtualenv in editable mode".format(
                cyan=Fore.CYAN,
                package_name=config.PACKAGE_NAME.get_value(),
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pip,
                "install",
                "--editable",
                config.dir_project_root,
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="Install the package to virtualenv in regular mode.",
    )
    def pip_install(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self.pip_uninstall(config, _dry_run=_dry_run, **kwargs)
        pgr_print(
            "{cyan}install '{package_name}' to virtualenv".format(
                cyan=Fore.CYAN,
                package_name=config.PACKAGE_NAME.get_value(),
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pip,
                "install",
                config.dir_project_root,
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="** Display useful information",
    )
    def info(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Display useful information, {green}path exists{cyan}, {red}path not exists".format(
                cyan=Fore.CYAN,
                green=Fore.GREEN,
                red=Fore.RED,
            )
        )
        print_path(
            "virtual environment directory",
            config.dir_venv,
        )
        print_path(
            "venv python executable path",
            config.path_venv_bin_python
        )
        print_path(
            "venv pip executable path",
            config.path_venv_bin_pip,
        )
        if IS_WINDOWS:
            print_line(
                "- {cyan}activate venv command: {path}".format(
                    cyan=Fore.CYAN,
                    path=colorful_path(config.path_venv_bin_activate),
                )
            )
        elif IS_MACOS or IS_LINUX:
            print_line(
                "- {cyan}activate venv command: source {path}".format(
                    cyan=Fore.CYAN,
                    path=colorful_path(config.path_venv_bin_activate),
                )
            )
        else:
            raise NotImplementedError

        print_line(
            "- {cyan}deactivate venv command: {reset}deactivate".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        print_path(
            "site-packages directory",
            config.dir_venv_site_packages,
        )
        print_path(
            "site-packages64 directory",
            config.dir_venv_site_packages_64,
        )

    @subcommand(
        help="** Install dev dependencies in requirements-dev.txt",
    )
    def req_dev(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}install dependencies in {reset}requirements-dev.txt to virtualenv".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pip,
                "install",
                "-r",
                config.path_requirements_dev_file,
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="Install doc dependencies in requirements-doc.txt",
    )
    def req_doc(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}install dependencies in {reset}requirements-doc.txt to virtualenv".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pip,
                "install",
                "-r",
                config.path_requirements_doc_file,
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="Install dev dependencies in requirements-test.txt",
    )
    def req_test(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}install dependencies in {reset}requirements-dev.txt to virtualenv".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pip,
                "install",
                "-r",
                config.path_requirements_test_file,
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="Display requirements file content",
    )
    def req_info(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}dependencies in {reset}requirements.txt".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        print()
        subprocess.call([
            "cat",
            config.path_requirements_file,
        ])
        print()

        pgr_print(
            "{cyan}dependencies in {reset}requirements-dev.txt".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        print()
        subprocess.call([
            "cat",
            config.path_requirements_dev_file,
        ])
        print()

        pgr_print(
            "{cyan}dependencies in {reset}requirements-doc.txt".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        print()
        subprocess.call([
            "cat",
            config.path_requirements_doc_file,
        ])
        print()

        pgr_print(
            "{cyan}dependencies in {reset}requirements-test.txt".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        print()
        subprocess.call([
            "cat",
            config.path_requirements_test_file,
        ])
        print()

    @subcommand(
        name="test-only",
        help="Run unit test with pytest.",
    )
    def test_pytest_only(self, config, _dry_run=False, **kwargs):  # pragma: no cover
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Run unit test in {reset}{dir_tests}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                dir_tests=config.dir_tests,
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pytest,
                config.dir_tests,
                "-s",
            ])
        pgr_print_done(indent=1)

    @subcommand(
        name="test",
        help="** Run unit test with pytest. Start over, reuse nothing.",
    )
    def test_pytest(self, config, _dry_run=False, **kwargs):  # pragma: no cover
        """
        :type config: RepoConfig
        """
        self.req_test(config, _dry_run=_dry_run, **kwargs)
        self.pip_dev_install(config, _dry_run=_dry_run, **kwargs)
        if _dry_run is False:
            remove_if_exists(config.dir_pytest_cache)
        self.test_pytest_only(config, _dry_run=_dry_run, **kwargs)

    @subcommand(
        name="cov-only",
        help="Run code coverage test in pytest.",
    )
    def test_cov_only(self, config, _dry_run=False, **kwargs):  # pragma: no cover
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Run code coverage test in {reset}{dir_tests}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                dir_tests=config.dir_tests,
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_pytest,
                config.dir_tests,
                "-s",
                "--cov={}".format(config.PACKAGE_NAME.get_value()),
                "--cov-report", "term-missing",
                "--cov-report", "annotate:{}".format(config.dir_coverage_annotate),
            ])
        pgr_print_done(indent=1)

    @subcommand(
        name="cov",
        help="** Run code coverage test in pytest. Start over, reuse nothing.",
    )
    def test_cov(self, config, _dry_run=False, **kwargs):  # pragma: no cover
        """
        :type config: RepoConfig
        """
        self.req_test(config, _dry_run=_dry_run, **kwargs)
        self.pip_dev_install(config, _dry_run=_dry_run, **kwargs)
        if _dry_run is False:
            remove_if_exists(config.dir_pytest_cache)
            remove_if_exists(config.dir_coverage_annotate)
        self.test_cov_only(config, _dry_run=_dry_run, **kwargs)

    @subcommand(
        name="tox-only",
        help="Run matrix test in tox with pytest",
    )
    def test_tox_only(self, config, _dry_run=False, **kwargs):  # pragma: no cover
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Run matrix test with tox settings in {reset}tox.ini".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                dir_tests=config.dir_tests,
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_tox,
                "--workdir \"{}\"".format(config.dir_project_root),
            ])
        pgr_print_done(indent=1)

    @subcommand(
        name="tox",
        help="** Run matrix test in tox with pytest. Start over, reuse nothing.",
    )
    def test_tox(self, config, _dry_run=False, **kwargs):  # pragma: no cover
        """
        :type config: RepoConfig
        """
        self.req_test(config, _dry_run=_dry_run, **kwargs)
        self.pip_dev_install(config, _dry_run=_dry_run, **kwargs)
        if _dry_run is False:
            remove_if_exists(config.dir_tox_cache)
        self.test_tox_only(config, _dry_run=_dry_run, **kwargs)

    @subcommand(
        name="pep8",
        help="Apply pep8 (https://www.python.org/dev/peps/pep-0008/) to source code and tests.",
    )
    def reformat_pep8_code_style(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self.req_dev(config, _dry_run=_dry_run)
        pgr_print(
            "{cyan}reformat python code style, execute {reset}{reformat_script}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                reformat_script=config.path_fix_code_style_script,
            )
        )
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_python,
                config.path_fix_code_style_script,
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="Build local documents with sphinx-doc.",
    )
    def build_doc_only(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Build doc at {reset}{doc_build_html}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                doc_build_html=config.dir_sphinx_doc_build_html,
            )
        )
        if _dry_run is False:
            subprocess.call([
                "make",
                "-C", config.dir_sphinx_doc,
                "html",
            ])
        pgr_print_done(indent=1)

    @subcommand(
        help="** Build local documents with sphinx-doc, start over, reuse nothing.",
    )
    def build_doc(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self.req_doc(config, _dry_run=_dry_run, **kwargs)
        self.pip_dev_install(config, _dry_run=_dry_run, **kwargs)
        if _dry_run is False:
            remove_if_exists(config.dir_sphinx_doc_build)
            remove_if_exists(os.path.join(
                config.dir_sphinx_doc_source, config.PACKAGE_NAME.get_value()))
        self.build_doc_only(config, _dry_run=_dry_run, **kwargs)

    @subcommand(
        help="Clear recently built local documents.",
    )
    def clean_doc(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Clean recently built doc at {reset}{doc_build}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                doc_build=config.dir_sphinx_doc_build,
            )
        )
        if _dry_run is False:
            remove_if_exists(config.dir_sphinx_doc_build)
        pgr_print_done(indent=1)

    @subcommand(
        help="** View recently build local documents.",
    )
    def view_doc(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Open recently built local html doc {reset}{doc_build_html}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                doc_build_html=config.dir_sphinx_doc_build_html,
            )
        )

        if os.path.exists(
            os.path.join(config.dir_sphinx_doc_build_html, "index.html")
        ):
            path_doc_index_html = os.path.join(config.dir_sphinx_doc_build_html, "index.html")
        elif os.path.exists(
            os.path.join(config.dir_sphinx_doc_build_html, "README.html")
        ):
            path_doc_index_html = os.path.join(config.dir_sphinx_doc_build_html, "README.html")
        else:
            raise ValueError

        if _dry_run is False:
            subprocess.call([OPEN_COMMAND, path_doc_index_html])

        pgr_print_done(indent=1)

    def _deploy_doc_to_s3(
        self,
        config,
        doc_version,
        s3_uri_doc_dir,
        doc_host_aws_profile,
        _dry_run=False,
        **kwargs
    ):
        """
        Deploy local html doc to S3.

        :type config: RepoConfig

        :type doc_version: str
        :param doc_version: "latest" or "versioned"

        :type s3_uri_doc_dir: str
        :type doc_host_aws_profile: str
        """
        ensure_s3_dir(s3_uri_doc_dir)

        pgr_print(
            "{cyan}deploy doc from local to s3 as {doc_version} doc ...".format(
                cyan=Fore.CYAN,
                doc_version=doc_version,
            )
        )

        pgr_print(
            "{cyan}{tab}aws s3 rm {reset}{s3_uri_doc_dir}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                tab=TAB,
                s3_uri_doc_dir=s3_uri_doc_dir,
            )
        )
        args = [
            "aws", "s3", "rm", s3_uri_doc_dir,
            "--recursive",
            "--only-show-errors",
        ]
        if doc_host_aws_profile is not None:
            args.extend(["--profile", doc_host_aws_profile])

        if _dry_run is False:
            subprocess.call(args)

        pgr_print(
            "{cyan}{tab}aws s3 sync {reset}{dir_sphinx_doc_build_html} {s3_uri_doc_dir}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                tab=TAB,
                dir_sphinx_doc_build_html=config.dir_sphinx_doc_build_html,
                s3_uri_doc_dir=s3_uri_doc_dir,
            )
        )
        args = [
            "aws", "s3", "sync",
            config.dir_sphinx_doc_build_html, s3_uri_doc_dir,
            "--only-show-errors",
        ]
        if doc_host_aws_profile is not None:
            args.extend(["--profile", doc_host_aws_profile])

        if _dry_run is False:
            subprocess.call(args)

        s3_console_url = s3_uri_to_url(s3_uri_doc_dir)
        pgr_print(
            "{cyan}{tab}view {doc_version} doc at {reset}{s3_console_url}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                tab=TAB,
                doc_version=doc_version,
                s3_console_url=s3_console_url,
            )
        )

        pgr_print_done(indent=1)

    @subcommand(
        help="Deploy local html doc to S3 as versioned document",
    )
    def deploy_doc_to_versioned(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self._deploy_doc_to_s3(
            config,
            doc_version="versioned",
            s3_uri_doc_dir=config.s3_uri_doc_dir_versioned,
            doc_host_aws_profile=config.DOC_HOST_AWS_PROFILE.get_value(),
            _dry_run=_dry_run,
            **kwargs
        )

    @subcommand(
        help="Deploy local html doc to S3 as latest document",
    )
    def deploy_doc_to_latest(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self._deploy_doc_to_s3(
            config,
            doc_version="latest",
            s3_uri_doc_dir=config.s3_uri_doc_dir_latest,
            doc_host_aws_profile=config.DOC_HOST_AWS_PROFILE.get_value(),
            _dry_run=_dry_run,
            **kwargs
        )

    @subcommand(
        help="Deploy local html doc to S3 as versioned document, and also as latest document optionally.",
    )
    def deploy_doc(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print("{cyan}deploy doc from local to s3 ...".format(cyan=Fore.CYAN))
        self.deploy_doc_to_versioned(config, _dry_run=_dry_run, **kwargs)

        pgr_print(
            "{cyan}also deploy latest doc (y/n)?".format(
                cyan=Fore.CYAN,
                tab=TAB,
            )
        )
        entered = str(input("")).lower().strip()
        if entered in ["y", "yes"]:
            self.deploy_doc_to_latest(config, _dry_run=_dry_run, **kwargs)
        else:
            pgr_print_done(indent=1)

    @subcommand(
        name="publish",
        help="** Publish this Package to https://pypi.org/.",
    )
    def publish_to_pypi(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self.req_dev(config, _dry_run=_dry_run, **kwargs)
        self.pip_dev_install(config, _dry_run=_dry_run, **kwargs)
        pgr_print(
            "{cyan}Publish {package_name} to {reset}https://pypi.org/project/{package_name}/".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                package_name=config.PACKAGE_NAME.get_value()
            )
        )
        if _dry_run is False:
            remove_if_exists(config.dir_pypi_build)
            remove_if_exists(config.dir_pypi_distribute)
            remove_if_exists(config.dir_pypi_egg)

        cwd = os.getcwd()
        os.chdir(config.dir_project_root)
        try:
            if _dry_run is False:
                subprocess.call([
                    config.path_venv_bin_python,
                    "setup.py",
                    "sdist",
                    "bdist_wheel",
                    "--universal",
                ])
                subprocess.call([
                    config.path_venv_bin_twine,
                    "upload",
                    "dist/*",
                ])
        except:
            pass
        os.chdir(cwd)
        pgr_print_done(indent=1)

    @subcommand(
        help="Run Jupyter notebook locally.",
    )
    def run_jupyter_notebook(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print("{cyan}run jupyter notebook ...".format(cyan=Fore.CYAN))
        if _dry_run is False:
            subprocess.call([
                config.path_venv_bin_jupyter,
                "notebook",
            ])

    # ==============================================================================
    # AWS Lambda Related
    # ==============================================================================
    @subcommand(
        help="Build AWS Lambda source code zip file.",
    )
    def build_lambda_source_code(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}build lambda source code at {reset}{path}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                path=config.path_lambda_build_source
            )
        )

        # filter all python source code need to add to the zip
        to_zip_list = list()
        for dirname, _, basename_list in os.walk(config.dir_python_lib):
            # ignore __pycache__
            if dirname.endswith("__pycache__"):
                continue
            for basename in basename_list:
                # ignore .pyc, .pyo,
                if basename.endswith(".pyc") or basename.endswith(".pyo"):
                    continue
                source_path = os.path.join(dirname, basename)
                archive_path = os.path.relpath(source_path, config.dir_project_root)
                to_zip_list.append((source_path, archive_path))

        if _dry_run is False:
            makedir_if_not_exists(config.dir_lambda_build)
            remove_if_exists(config.path_lambda_build_source)

            with ZipFile(config.path_lambda_build_source, "w") as f:
                for source_path, archive_path in to_zip_list:
                    f.write(source_path, archive_path)

        pgr_print_done(indent=1)

    def _upload_lambda_zip(
        self,
        config,
        source_or_layer,
        path,
        _dry_run=False,
        **kwargs
    ):
        """
        :type config: RepoConfig
        :param source_or_layer: "source code" or "layer"
        """
        pgr_print(
            "{cyan}upload lambda {source_or_layer} from {reset}{path} {cyan}to AWS S3".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                source_or_layer=source_or_layer,
                path=path,
            )
        )
        if source_or_layer == "source code":
            s3_uri_lambda_deploy_versioned_dir = config.s3_uri_lambda_deploy_versioned_source_dir
        elif source_or_layer == "layer":
            s3_uri_lambda_deploy_versioned_dir = config.s3_uri_lambda_deploy_versioned_layer_dir
        else:
            raise ValueError

        if os.path.exists(path):
            source_md5 = fingerprint.of_file(path)
            bucket, prefix = split_s3_uri(s3_uri_lambda_deploy_versioned_dir)
            key = s3_key_smart_join(
                parts=[prefix, "{}.zip".format(source_md5)],
                is_dir=False,
            )
            s3_uri = join_s3_uri(bucket, key)
            pgr_print(
                "{cyan}{tab}upload to {reset}{s3_uri}".format(
                    cyan=Fore.CYAN,
                    tab=TAB,
                    reset=Style.RESET_ALL,
                    s3_uri=s3_uri,
                )
            )
            args = [
                "aws", "s3", "cp",
                path, s3_uri,
            ]
            aws_profile = config.AWS_LAMBDA_DEPLOY_AWS_PROFILE.get_value()
            if aws_profile is not None:
                args.extend(["--profile", aws_profile])
            if _dry_run is False:
                subprocess.call(args)
            pgr_print_done(indent=1)
        else:
            pgr_print(
                "{red}{tab}{path} {cyan}not found!".format(
                    tab=TAB,
                    red=Fore.RED,
                    cyan=Fore.CYAN,
                    path=path,
                )
            )

    @subcommand(
        help="Upload AWS Lambda source code zip file to S3.",
    )
    def upload_lambda_source_code(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self._upload_lambda_zip(
            config,
            source_or_layer="source code",
            path=config.path_lambda_build_source,
            _dry_run=_dry_run,
            **kwargs
        )

    def _find_docker(self):
        """
        Find the absolute path of the docker executable
        """
        if IS_MACOS or IS_LINUX:
            locations = [
                os.path.join("/", "usr", "bin", "docker"),
                os.path.join("/", "usr", "local", "bin", "docker"),
            ]
            for p in locations:
                if os.path.exists(p):
                    return p
            raise EnvironmentError("We cannot find docker, are you sure docker is installed?")
        else:
            raise EnvironmentError("We can only find docker bin on MacOS or Linux!")

    @subcommand(
        help="Build lambda layer using a AWS lambda runtime compatible docker image.",
    )
    def build_lambda_layer(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        config.ensure_attr_not_none(config.AWS_LAMBDA_BUILD_DOCKER_IMAGE.name)
        config.ensure_attr_not_none(config.AWS_LAMBDA_BUILD_DOCKER_IMAGE_WORKSPACE_DIR.name)
        pgr_print(
            "{cyan}build lambda layer in {reset}{docker_image} {cyan}docker container at {reset}{path}".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
                docker_image=config.AWS_LAMBDA_BUILD_DOCKER_IMAGE.get_value(),
                path=config.path_lambda_build_layer
            )
        )
        makedir_if_not_exists(config.dir_lambda_build)

        path_bin_docker = self._find_docker()
        container_only_build_lbd_layer_script_path = os.path.join(
            config.AWS_LAMBDA_BUILD_DOCKER_IMAGE_WORKSPACE_DIR.get_value(),
            "bin",
            "container-only-build-lambda-layer.sh",
        )

        if _dry_run is False:
            subprocess.call([
                path_bin_docker, "run",
                "-v", "{}:{}".format(
                    config.dir_project_root,
                    config.AWS_LAMBDA_BUILD_DOCKER_IMAGE_WORKSPACE_DIR.get_value(),
                ),
                "--rm", config.AWS_LAMBDA_BUILD_DOCKER_IMAGE.get_value(),
                "bash", container_only_build_lbd_layer_script_path,
            ])

    @subcommand(
        help="Upload AWS Lambda layer zip file to S3.",
    )
    def upload_lambda_layer(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self._upload_lambda_zip(
            config,
            source_or_layer="layer",
            path=config.path_lambda_build_layer,
            _dry_run=_dry_run,
            **kwargs
        )

    @subcommand(
        help="Deploy recently built AWS lambda layer.",
    )
    def deploy_lambda_layer(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}deploy a new version of lambda layer from AWS S3".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        if os.path.exists(config.path_lambda_build_layer):
            source_md5 = fingerprint.of_file(config.path_lambda_build_layer)
            bucket, prefix = split_s3_uri(config.s3_uri_lambda_deploy_versioned_layer_dir)
            key = s3_key_smart_join(
                parts=[prefix, "{}.zip".format(source_md5)],
                is_dir=False,
            )
            s3_console_url = make_s3_console_url(bucket=bucket, prefix=key)
            pgr_print(
                "{cyan}{tab}run {reset}'aws lambda publish-layer-version' {cyan}command, preview layer file at {reset}{s3_console_url}".format(
                    cyan=Fore.CYAN,
                    tab=TAB,
                    reset=Style.RESET_ALL,
                    s3_console_url=s3_console_url,
                )
            )
            args = [
                "aws", "lambda", "publish-layer-version",
                "--layer-name", config.aws_lambda_layer_name,
                "--description",
                "dependency layer for all functions in '{}' project".format(config.aws_lambda_layer_name),
                "--content", "S3Bucket={},S3Key={}".format(
                    config.AWS_LAMBDA_DEPLOY_S3_BUCKET.get_value(), key,
                ),
                "--compatible-runtimes", "python{}.{}".format(
                    config.DEV_PY_VER_MAJOR.get_value(),
                    config.DEV_PY_VER_MINOR.get_value()
                ),
            ]
            aws_profile = config.AWS_LAMBDA_DEPLOY_AWS_PROFILE.get_value()
            if aws_profile is not None:
                args.extend(["--profile", aws_profile])
            if _dry_run is False:
                subprocess.call(args)
            pgr_print(
                "{cyan}{tab}open {reset}{url} {cyan}to view layer".format(
                    cyan=Fore.CYAN,
                    tab=TAB,
                    reset=Style.RESET_ALL,
                    url=config.url_lambda_layer_console,
                )
            )
            pgr_print_done(indent=1)
        else:
            pgr_print(
                "{red}{tab}{path} {cyan}not found!".format(
                    tab=TAB,
                    red=Fore.RED,
                    cyan=Fore.CYAN,
                    path=config.path_lambda_build_layer,
                )
            )

    @subcommand(
        name="bud-lambda-layer",
        help="** Build, upload and deploy a new AWS lambda layer.",
    )
    def build_upload_deploy_lambda_layer(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        self.build_lambda_layer(config, _dry_run=_dry_run)
        self.upload_lambda_layer(config, _dry_run=_dry_run, **kwargs)
        self.deploy_lambda_layer(config, _dry_run=_dry_run, **kwargs)

    def _ensure_chalice_lambda_app_dir(self, config):
        """
        :type config: RepoConfig
        """
        files = [
            config.path_aws_chalice_config_json,
            config.path_aws_chalice_app_py,
        ]
        for p in files:
            if not os.path.exists(p):
                raise ValueError("{} not exists".format(p))

    @subcommand(
        help="** Deploy AWS Lambda app with AWS Chalice.",
    )
    def chalice_deploy(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Deploy AWS Lambda app with AWS Chalice".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )

        self._ensure_chalice_lambda_app_dir(config)

        # reset the ``vendor`` dir
        remove_if_exists(config.dir_aws_chalice_vendor)
        makedir_if_not_exists(config.dir_aws_chalice_vendor)

        # copy source code to ``vendor`` dir
        copy_python_code(
            config.dir_python_lib,
            config.dir_aws_chalice_vendor_source,
        )

        # invoke chalice cli
        args = [
            config.path_venv_bin_chalice,
            "--project-dir", config.dir_lambda_app,
            "deploy",
        ]
        args.extend(kwargs["_args"])
        aws_profile = config.AWS_LAMBDA_DEPLOY_AWS_PROFILE.get_value()
        if aws_profile is not None:
            args.extend(["--profile", aws_profile])

        if _dry_run is False:
            subprocess.call(args)

    @subcommand(
        help="** Delete AWS Lambda app with AWS Chalice.",
    )
    def chalice_delete(self, config, _dry_run=False, **kwargs):
        """
        :type config: RepoConfig
        """
        pgr_print(
            "{cyan}Delete AWS Lambda app with AWS Chalice".format(
                cyan=Fore.CYAN,
                reset=Style.RESET_ALL,
            )
        )
        self._ensure_chalice_lambda_app_dir(config)

        args = [
            config.path_venv_bin_chalice,
            "--project-dir", config.dir_lambda_app,
            "delete",
        ]
        args.extend(kwargs["_args"])
        aws_profile = config.AWS_LAMBDA_DEPLOY_AWS_PROFILE.get_value()
        if aws_profile is not None:
            args.extend(["--profile", aws_profile])

        if _dry_run is False:
            subprocess.call(args)


actions = Actions()
